"""
Example to generate multichannel babble speech with a desired spatial coherence.
"""

import numpy as np
from scipy.io import wavfile
import matplotlib.pyplot as plt
import anf_generator as anf


def plot_spatial_coherence(coherence_gen, coherence_target, params):
    num_channels = len(params.mic_positions)
    sample_frequency = params.sample_frequency
    nfft = params.nfft

    # Plot example of generated vs target spatial coherence
    freqs = np.linspace(0, sample_frequency / 2, nfft // 2 + 1) / 1000  # Freq. in kHz
    if params.sc_type in ["spherical", "cylindrical"]:
        fig = plt.figure()
        plt.subplots_adjust(hspace=0.5)
        fig.suptitle("Spatial Coherence")

        for m in range(num_channels - 1):
            plt.subplot(num_channels - 1, 1, m + 1)
            plt.plot(freqs, np.real(coherence_gen[0, m + 1, :]), "-k", linewidth=2)
            plt.plot(freqs, np.real(coherence_target[0, m + 1, :]), "-.b", linewidth=2)
            plt.title(f"Sensors 1-{m+2}")
            plt.axis([0, sample_frequency / 2000, -1, 1])
            plt.grid(True)
            plt.ylabel("Spatial coherence")

        fig.supxlabel("Frequency [kHz]")
    else:
        fig = plt.figure(constrained_layout=True)
        fig.suptitle("Spatial Coherence")

        subfigs = fig.subfigures(nrows=num_channels - 1, ncols=1)

        for row, subfig in enumerate(subfigs):
            subfig.suptitle(f"Sensors 1-{row+2}")

            axs = subfig.subplots(nrows=1, ncols=2)
            axs[0].plot(freqs, np.real(coherence_gen[0, row + 1, :]), "-k", linewidth=2)
            axs[0].plot(
                freqs, np.real(coherence_target[0, row + 1, :]), "-.b", linewidth=2
            )
            axs[0].grid(True)
            axs[0].set_ylim([-1, 1])
            axs[0].set_title("Real")

            axs[1].plot(freqs, np.imag(coherence_gen[0, row + 1, :]), "-k", linewidth=2)
            axs[1].plot(
                freqs, np.imag(coherence_target[0, row + 1, :]), "-.b", linewidth=2
            )
            axs[1].grid(True)
            axs[1].set_ylim([-1, 1])
            axs[1].set_title("Imaginary")

            if row == num_channels - 2:
                axs[0].set_xlabel("Frequency [kHz]")
                axs[1].set_xlabel("Frequency [kHz]")

    return


def main():
    # Set decomposition and processing methods
    decomposition = "evd"  # Type of decomposition: 'evd' or 'chd'
    processing = (
        "balance+smooth"  # Processing: 'standard', 'smooth', 'balance', 'balance+smooth'
    )
    duration = 10  # Input duration in seconds
    num_channels = 4  # Number of microphones
    evaluate = True # Set to True to evaluate results

    # Set parameters to define the target coherence
    params = anf.CoherenceMatrix.Parameters(
        mic_positions=np.array([[0.04 * i, 0, 0] for i in range(num_channels)]),
        # Type of Spatial Coherence: 'corcos', 'spherical', or 'cylindrical'
        sc_type="spherical",
        sample_frequency=8000,
        nfft=1024,
    )

    # Summarize main parameters
    print(f"Number of channels: {len(params.mic_positions)}")
    print(f"Spatial coherence: {params.sc_type}")
    print(f"Decomposition: {decomposition}")
    print(f"Processing: {processing}\n")

    # Read the WAV file
    sample_frequency_wav, input_signal = wavfile.read('babble_8kHz.wav')
    if sample_frequency_wav != params.sample_frequency:
        raise ValueError("The sample frequency of the input signal does not "
                         "match the desired sampling frequency.")

    # Calculate the number of samples
    num_samples = duration * sample_frequency_wav

    # Ensure the input signal is long enough
    if len(input_signal) < num_channels * num_samples:
        raise ValueError(f"Input signal is not long enough to extract {num_channels} "
                         f"non-overlapping segments of {duration} seconds each.")

    # Slice the input signal into non-overlapping segments
    trimmed_signal = input_signal[:num_channels * num_samples]
    input_signals = trimmed_signal.reshape(num_channels, num_samples)

    # Generate output signals with the desired spatial coherence
    output_signals, coherence_target, mixing_matrix = anf.generate_signals(
        input_signals, params, decomposition, processing)

    # Scale and save the generated signals to a WAV file
    scaled_signals = (output_signals / np.max(np.abs(output_signals)) * 32767).astype(np.int16)
    wavfile.write('mc_babble_8kHz.wav', params.sample_frequency, scaled_signals.T)

    if evaluate:
        # Evaluate mixing matrix
        eval_results = mixing_matrix.evaluate_matrix(plot=True)

        # Estimate coherence from the generated sensor signals
        coherence_gen = anf.estimate_coherence(output_signals, params.nfft)

        # Plot spatial coherence matrices
        plot_spatial_coherence(coherence_gen, coherence_target.matrix, params)

        # Compute error between target and generated coherence
        xi = np.sum(np.abs(coherence_gen - coherence_target.matrix) ** 2, axis=(0, 1))
        xi_avg = 10 * np.log10(np.average(xi))

        # Summarize evaluation results
        print("Improvements:")
        print(
            f'Spectral Variation: '
            f'{round(eval_results["smoothness"] - eval_results["smoothness_std"],2)} dB'
        )
        print(
            f'Mix Balance: '
            f'{round(eval_results["balance"] - eval_results["balance_std"],2)} dB'
        )
        print(
            f'Coherence Error: '
            f'{round(eval_results["coh_error"] - eval_results["coh_error_std"],2)} dB'
        )
        print(f"Average Coherence Error: "
              f"{round(xi_avg, 2)} dB\n")
        
        plt.show()


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'anf-generator'
copyright = '2024, Emanuel Habets'
author = 'Emanuel Habets'

# The full version, including alpha/beta/rc tags
release = '0.1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.imgmath",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinxcontrib.bibtex",
    "numpydoc",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', "**tests**", 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Options for LaTeX output ---------------------------------------------

# latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
# }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#     (
#         "index",
#         "anf_generator.tex",
#         "anf_generator Documentation",
#         "Emanuel Habets",
#         "manual",
#     ),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#     ("index", "anf_generator", "anf_generator Documentation", ["Emanuel Habets"], 1)
# ]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = [
#     (
#         "index",
#         "anf_generator",
#         "anf_generator Documentation",
#         "Emanuel Habets",
#         "anf_generator",
#         "Generating coherence-constrained multisensor signals.",
#         "Miscellaneous",
#     ),
# ]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

bibtex_bibfiles = ["refs.bib"]



import numpy as np
import scipy.linalg
from . import MixingMatrix
from . import CoherenceMatrix

def generate_signals(input_signals, params, decomposition='evd', processing='balance+smooth'):
    """
    Generate M output signals with the desired spatial coherence.
    
    Input:
        input_signals: M mutually independent input signals [Channels x Samples]
        decomposition: Type of decomposition: 'evd' or 'chd'
        processing: Type of processing: 'standard', 'smooth', 'balance', 'balance+smooth'
        params: Parameters object
    
    Output:
        output_signals: Generated output signals [Channels x Samples]
        coherence_target: Target coherence matrix
        mixing_matrix: Mixing matrix object
    """

    # Generate target spatial coherence
    coherence_target = CoherenceMatrix.CoherenceMatrix(params)

    # Compute mixing matrix
    mixing_matrix = MixingMatrix.MixingMatrix(
        coherence_matrix=coherence_target,
        decomposition=decomposition,
        processing=processing
    )

    # Generate sensor signals with target spatial coherence
    output_signals = mix_signals(input_signals, mixing_matrix)

    return output_signals, coherence_target, mixing_matrix


def mix_signals(y, mixing_matrix: MixingMatrix.MixingMatrix):
    """
    Mix M mutually independent signals such that the mixed signals
    exhibit a specific spatial coherence.

    Parameters:
        n (numpy.array): M signals in the time domain [Channels x Samples]
        mixing_matrix (MixingMatrix): Mixing matrix object

    Returns:
        x (numpy.array): M generated signals [Channels x Samples]
    """

    # Initialization
    L = y.shape[1]  # Length input signal
    K = mixing_matrix.coherence_matrix.params.nfft  # FFT length

    # Compute STFT of the input signals
    _, _, Y = scipy.signal.stft(y, fs=1.0, window='hann', nperseg=K, noverlap=K//4, nfft=K,
                                return_onesided=True, boundary='zeros', padded=True, axis=-1,
                                scaling='spectrum')
    Y = np.transpose(Y, axes=[0, 2, 1]) # Transpose to [Channels x Samples x Frequencies]

    # Initialize STFT output
    X = np.zeros_like(Y)

    # Filter input signals
    for k in range(K//2, 0, -1):
        X[:, :, k] = np.conj(mixing_matrix.matrix[:, :, k]).T @ Y[:, :, k]

    # Inverse STFT
    X = np.transpose(X, axes=[0, 2, 1]) # Transpose to [Channels x Frequencies x Samples]
    _, x = scipy.signal.istft(X, fs=1.0, window='hann', nperseg=K, noverlap=K//4, nfft=K,
                              input_onesided=True, time_axis=-1, freq_axis=-2, scaling='spectrum')
    x = x[:, :L]  # Output signals

    return x


def estimate_coherence(x, nfft):
    """
    Compute the complex spatial coherence of x using the Welch's periodogram approach.

    Parameters:
        x (numpy.array): Multi-channel time signals [Channels x Samples]
        nfft (int): FFT length

    Returns:
        CC (numpy.array): Estimated complex coherence matrix [Channels x Channels x Frequencies]
    """
    if x.shape[1] == 1:
        raise ValueError(
            'Number of channels must be > 1 to compute the spatial coherence.')

    # Compute STFT with 75% overlap of the input
    _, _, X = scipy.signal.stft(x, fs=1.0, window='hann', nperseg=nfft, noverlap=nfft//4, 
                                nfft=nfft, return_onesided=True, boundary='zeros', 
                                padded=True, axis=-1, scaling='spectrum')

    M, K, L = X.shape  # Number of channels, frequencies, and frames

    # Compute PSD matrix
    psd_matrix = np.zeros((M, M, K), dtype=np.complex_)
    for l in range(L):
        # Compute instantaneous spectrum for the l-th frame
        Xf = X[:, :, l].T

        # Compute instantaneous PSD matrix
        XX = np.einsum('ij,ik->jki', Xf, np.conj(Xf))

        # Update
        psd_matrix += XX

    # Compute complex coherence matrix
    CC = np.zeros((M, M, K), dtype=np.complex_)
    for r in range(M):
        for c in range(M):
            CC[r, c, :] = psd_matrix[r, c, :] / \
                np.sqrt(psd_matrix[r, r, :] * psd_matrix[c, c, :])

    return CC



import numpy as np
import scipy.linalg
import matplotlib.pyplot as plt
from . import CoherenceMatrix


class MixingMatrix:
    def __init__(
        self, coherence_matrix: CoherenceMatrix.CoherenceMatrix, decomposition: str, processing: str
    ):
        self.coherence_matrix = coherence_matrix
        self.decomposition = decomposition
        self.processing = processing

        self.matrix, self.matrix_std = self.compute_mixing_matrix()

    def compute_mixing_matrix(self):
        # Initialization
        M = self.get_num_channels()  # Number of channels
        K = self.get_nfft()  # Number of frequency bins
        C = np.zeros((M, M, K), dtype=np.complex_)  # STFT mixing matrix

        # Direct Current component definition for the mixing matrix (fully coherent)
        C[:, :, 0] = np.ones((M, M)) / np.sqrt(M)

        # Generate mixing matrix in the STFT domain for each frequency bin k
        DC = self.coherence_matrix.matrix  # Target coherence matrix
        for k in range(K // 2, 0, -1):
            # Rank of the target coherence matrix
            DC_rank = np.linalg.matrix_rank(DC[:, :, k])

            if self.decomposition == "chd" and DC_rank == M:
                C[:, :, k] = scipy.linalg.cholesky(DC[:, :, k], lower=False)
            elif self.decomposition == "evd" or (
                self.decomposition == "chd" and DC_rank < M
            ):
                D, V = np.linalg.eigh(DC[:, :, k])
                D = np.clip(D, 0, np.inf)
                # np.dot(np.diag(np.sqrt(D)),V.conj().T)
                C[:, :, k] = np.sqrt(D)[:, np.newaxis] * V.conj().T
            else:
                raise ValueError(
                    'Unknown decomposition method specified. Please select "CHD" or "EVD".'
                )

        # Save original CHD/EVD mixing matrix for evaluation
        C_std = np.copy(C)

        # Apply selected processing method
        for k in range(K // 2, 0, -1):
            if self.processing == "debug":  # Only for testing STFT and iSTFT
                C[:, :, k] = np.eye(M)
            elif self.processing == "standard":
                continue
            elif self.processing == "balance":
                C[:, :, k] = MixingMatrix.balance(C[:, :, k])
            elif self.processing == "smooth":
                C[:, :, k - 1] = MixingMatrix.smooth(C[:, :, k], C[:, :, k - 1])
            elif self.processing == "balance+smooth":
                if k == K // 2:
                    C[:, :, k] = MixingMatrix.balance(C[:, :, k], U_type="orthogonal")
                C[:, :, k - 1] = MixingMatrix.smooth(C[:, :, k], C[:, :, k - 1])
                C[:, :, k - 1] = MixingMatrix.balance_perserving_smoothness(
                    C[:, :, k - 1]
                )
            else:
                raise ValueError(
                    'Unknown processing method specified. Please select "standard", "balance", "smooth", or "balance+smooth".'
                )

        return C, C_std

    def evaluate_matrix(self, plot=True):
        M = self.get_num_channels()  # Number of channels
        K = self.get_nfft()

        C_std = self.matrix_std[:, :, 0 : K // 2]
        C = self.matrix[:, :, 0 : K // 2]

        # Compute frequency-wise balance
        bal_std = np.sum(np.abs(C_std), axis=(0, 1)) / (M * np.sqrt(M))
        bal = np.sum(np.abs(C), axis=(0, 1)) / (M * np.sqrt(M))

        # Compute mean balance in dB
        bal_std_mean = 20 * np.log10(np.mean(bal_std))
        bal_mean = 20 * np.log10(np.mean(bal))

        # Compute frequency-wise smoothness
        smooth_std = np.sum(np.abs(np.diff(C_std, axis=2)) ** 2, axis=(0, 1))
        smooth = np.sum(np.abs(np.diff(C, axis=2)) ** 2, axis=(0, 1))

        # Compute mean smoothness in dB
        smooth_std_mean = 10 * np.log10(np.mean(smooth_std))
        smooth_mean = 10 * np.log10(np.mean(smooth))

        # Compute coherence error and IRs
        xi_std, xi, c_std, c = self.coherence_error()

        eval_metrics = {
            "balance_std": bal_std_mean,
            "smoothness_std": smooth_std_mean,
            "balance": bal_mean,
            "smoothness": smooth_mean,
            "coh_error_std": xi_std,
            "coh_error": xi,
        }

        if plot:
            self.__plot_smoothness_balance(
                smooth, smooth_std, bal, bal_std, eval_metrics
            )
            self.__plot_filters(c_std, c)

        return eval_metrics

    def coherence_error(self, plot=True):
        C1_std = self.matrix_std
        C1 = self.matrix

        K1 = self.get_nfft()  # Current DFT length
        K2 = int(np.floor(np.pi * K1))  # Increased DFT length K2 > K1
        M = self.get_num_channels()  # Number of channels

        # Pre-allocate memory
        c1_std = np.zeros((M, M, K1), dtype=np.complex_)
        c1 = np.zeros((M, M, K1), dtype=np.complex_)
        C2_std = np.zeros((M, M, K2), dtype=np.complex_)
        C2 = np.zeros((M, M, K2), dtype=np.complex_)

        for p in range(M):
            for q in range(M):
                # Inverse DFT to get IR
                c1_std[p, q, :] = np.fft.irfft(C1_std[p, q, :], n=K1)
                c1[p, q, :] = np.fft.irfft(C1[p, q, :], n=K1)

                # Circular shift
                c1_std[p, q, :] = np.roll(c1_std[p, q, :], K1 // 2)
                c1[p, q, :] = np.roll(c1[p, q, :], K1 // 2)

                # DFT with increased frame length K2
                C2_std[p, q, :] = np.fft.fft(c1_std[p, q, :], n=K2)
                C2[p, q, :] = np.fft.fft(c1[p, q, :], n=K2)

        # Generate target spatial coherence with DFT length K2 > K1
        params2 = self.coherence_matrix.params.copy()
        params2.nfft = K2
        DC2 = CoherenceMatrix.CoherenceMatrix(params2)

        # Compute generated coherence with DFT-length K2 as C'*C and nMSE
        G_std = np.zeros((M, M, K2 // 2 + 1), dtype=np.complex_)
        G = np.zeros((M, M, K2 // 2 + 1), dtype=np.complex_)
        nMSEk_std = np.zeros(K2 // 2 + 1)
        nMSEk = np.zeros(K2 // 2 + 1)
        for k in range(K2 // 2 + 1):
            G_std[:, :, k] = C2_std[:, :, k].conj().T @ C2_std[:, :, k]
            G[:, :, k] = C2[:, :, k].conj().T @ C2[:, :, k]
            nMSEk_std[k] = (
                np.linalg.norm(G_std[:, :, k] - DC2.matrix[:, :, k], "fro") ** 2
            )
            nMSEk[k] = np.linalg.norm(G[:, :, k] - DC2.matrix[:, :, k], "fro") ** 2

        # Compute normalized mean squared error between target and generated coherence
        xi_std = 10 * np.log10(np.mean(nMSEk_std))
        xi = 10 * np.log10(np.mean(nMSEk))

        if plot:
            self.__plot_coherence(G, G_std, DC2.matrix, xi, xi_std)

        return xi_std, xi, c1_std, c1

    def get_nfft(self):
        return self.coherence_matrix.params.nfft

    def get_num_channels(self):
        return len(self.coherence_matrix.params.mic_positions)

    def __plot_smoothness_balance(self, smooth, smooth_std, bal, bal_std, eval_metrics):
        # Define labels for ledgend
        smoothness_label = (f'{self.decomposition} {self.processing}: \n'
                            f'ε = {eval_metrics["smoothness"]:.1f} dB, '
                            f'β = {eval_metrics["balance"]:.1f} dB')
        smoothness_std_label = (f'{self.decomposition}: \n'
                                f'ε = {eval_metrics["smoothness_std"]:.1f} dB, '
                                f'β = {eval_metrics["balance_std"]:.1f} dB')

        # Plot Spectral Smoothness and Mix Balance
        plt.figure()

        plt.subplot(2, 1, 1)
        plt.plot(smooth, linewidth=2, label=smoothness_label)
        plt.plot(smooth_std, linestyle="-.", linewidth=2, label=smoothness_std_label)
        plt.legend()
        plt.grid(True)
        plt.ylabel("Smoothness")
        plt.ylim(0, np.amax(np.concatenate((smooth, smooth_std))))

        plt.subplot(2, 1, 2)
        plt.plot(bal, linewidth=2)
        plt.plot(bal_std, linestyle="-.", linewidth=2)
        plt.grid(True)
        plt.xlabel("Frequency")
        plt.ylabel("Balance")
        plt.ylim(0, 1)

    def __plot_coherence(self, G, G_std, DC2, xi, xi_std):
        K2 = DC2.shape[2]  # Increased DFT length
        sample_frequency = (
            self.coherence_matrix.params.sample_frequency
        )  # Sample frequency
        # Frequency vector in Hz
        Freqs = np.linspace(0, sample_frequency / 2, K2)

        # Procesd the decomposition and processing strings
        decomposition = self.decomposition
        processing = self.processing

        fig = plt.figure()
        plt.subplots_adjust(hspace=0.6)

        if (
            self.coherence_matrix.params.sc_type == "spherical"
            or self.coherence_matrix.params.sc_type == "cylindrical"
        ):
            # Case diffuse (real-valued coherence)
            plt.subplot(2, 1, 1)
            plt.plot(Freqs / 1000, np.real(G_std[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.real(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.title(decomposition)
            plt.grid(True)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.ylabel("Spatial coherence")
            plt.legend([rf"$\xi$ = {xi_std:.1f} dB", "Target"])

            plt.subplot(2, 1, 2)
            plt.plot(Freqs / 1000, np.real(G[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.real(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.title(f"{decomposition} {processing}")
            plt.grid(True)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.ylabel("Spatial coherence")
            plt.legend([rf"$\xi$ = {xi:.1f} dB", "Target"])

            fig.suptitle("Coherence error - Sensors 1-2")
            fig.supxlabel("Frequency [kHz]")
        else:
            # Case 'corcos' or generally complex-valued coherence
            plt.subplot(2, 2, 1)
            plt.plot(Freqs / 1000, np.real(G_std[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.real(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.title(decomposition)
            plt.grid(True)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.ylabel("Real")
            plt.legend([rf"$\xi$ = {xi_std:.1f} dB", "Target"])

            plt.subplot(2, 2, 2)
            plt.plot(Freqs / 1000, np.imag(G_std[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.imag(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.ylabel("Imag")
            plt.grid(True)

            plt.subplot(2, 2, 3)
            plt.plot(Freqs / 1000, np.real(G[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.real(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.title(f"{decomposition} {processing}")
            plt.grid(True)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.legend([rf"$\xi$ = {xi:.1f} dB", "Target"])

            plt.subplot(2, 2, 4)
            plt.plot(Freqs / 1000, np.imag(G[1, 2, :]), "-k", linewidth=2)
            plt.plot(Freqs / 1000, np.imag(DC2[1, 2, :]), "-.b", linewidth=2)
            plt.xlim([0, sample_frequency / 2000])
            plt.ylim([-1, 1])
            plt.grid(True)

            fig.suptitle("Coherence error - Sensors 1-2")
            fig.supylabel("Spatial Coherence")
            fig.supxlabel("Frequency [kHz]")

        # plt.show()

    def __plot_filters(self, c_std, c):
        # Choose which IRs of the mixing matrix to plot
        row = 1
        column = 1

        # Normalize IRs
        c_std_normalized = np.abs(c_std[row, column, :]) / np.max(
            np.abs(c_std[row, column, :])
        )
        c_normalized = np.abs(c[row, column, :]) / np.max(np.abs(c[row, column, :]))

        # Plot IRs of the mixing matrix
        fig = plt.figure()

        plt.subplot(1, 2, 1)
        plt.plot(20 * np.log10(c_std_normalized), linewidth=1)
        plt.title(f"{self.decomposition} - IR({row+1},{column+1})")
        plt.grid(True)
        plt.xlabel("Time")

        plt.subplot(1, 2, 2)
        plt.plot(20 * np.log10(c_normalized), linewidth=1)
        plt.title(f"{self.decomposition} {self.processing} - IR({row+1},{column+1})")
        plt.grid(True)
        plt.xlabel("Time")

        fig.supylabel("Magnitude [dB]")

    @staticmethod
    def balance(C, U_type="unitary"):
        """
        Compute optimal balance B = U * C with max l1(U * C) for U
        unitary or orthogonal.

        Parameters:
            C (numpy.array): Mixing matrix [Channels x Channels]
            type: 'unitary' (default) for U unitary, and 'orthogonal' for U orthogonal

        Returns:
            B (numpy.array): balance mixing matrix [Channels x Channels]
        """
        num_channels = C.shape[0]  # Number of channels

        # Init with a matrix of scaled ones
        H = np.ones((num_channels, num_channels)) / np.sqrt(num_channels)

        # Perform optimization
        if U_type == "unitary":
            U = MixingMatrix.absoluteUnitaryProcrustes(C, H)
        elif U_type == "orthogonal":
            U = MixingMatrix.absoluteOrthogonalProcrustes(C, H)

        # Compute balanced matrix
        B = U @ C

        return B

    @staticmethod
    def absoluteUnitaryProcrustes(A, B):
        """
        Solves || abs(U * A) - abs(B) ||_F with unitary U.
        """
        MaximumTrails = 10  # Number of trials
        N = A.shape[0]  # Number of channels
        newPhases = np.exp(1j * np.angle(B, deg=False))  # Phase matrix of B

        bestMatrix = None
        bestL1 = -np.inf

        for _ in range(MaximumTrails):
            # Perform optimization
            U = MixingMatrix.signVariableExchange(A, newPhases, False)

            # Evaluate l1-norm of obtained unitary matrix
            l1 = MixingMatrix.l1norm(U @ A)

            # Check if this is the best matrix so far
            if l1 > bestL1:
                bestL1 = l1
                bestMatrix = U

            # Generate new (random) normalized phase matrix
            newPhases = np.exp(1j * np.random.rand(N, N) * 2 * np.pi)

        return bestMatrix

    @staticmethod
    def absoluteOrthogonalProcrustes(A, B):
        """
        Solves || abs(U * A) - abs(B) ||_F with ortogonal U for
        A, B and U being real-valued.
        """
        MaximumTrails = 10  # Number of trials
        N = A.shape[0]  # Number of channels
        newSigns = np.sign(np.real(B) + np.finfo(float).eps)  # Sign matrix of B

        bestMatrix = None
        bestL1 = -np.inf

        for _ in range(MaximumTrails):
            # Perform optimization
            U = MixingMatrix.signVariableExchange(A, newSigns, False)

            # Evaluate l1-norm of obtained unitary matrix
            l1 = MixingMatrix.l1norm(U @ A)

            # Check if this is the best matrix so far
            if l1 > bestL1:
                bestL1 = l1
                bestMatrix = U

            # Generate new (random) normalized phase matrix
            newSigns = np.sign(np.random.rand(N, N) * 2 - 1)

        return bestMatrix

    @staticmethod
    def signVariableExchange(A, B, verbose=False):
        """
        Perform the variable exchange optimization.

        Parameters:
            A (numpy.array): First matrix
            B (numpy.array): Second matrix (Phase matrix)
            verbose (bool): If True, print all messages. If False, print only warning messages.

        Returns:
            U (numpy.array): Unitary matrix that minimizes (U*A - B) in the Frobenius norm
        """
        maxIter = 1000  # Max number of iterations
        bestFit = np.inf  # Best fit in the Frobenius-norm sense
        Phases = np.exp(1j * np.angle(B))  # Phase matrix of B

        newFit = np.zeros(maxIter)
        l1Fit = np.zeros(maxIter)
        # Us = []

        for counter in range(maxIter):
            # New phase matrix
            phaseB = Phases

            # Procrustes solution
            U = MixingMatrix.procrustes(A, phaseB)

            # Compute Frobenius norm of (U*A - phaseB)
            newFit[counter] = np.linalg.norm((U @ A) - phaseB, "fro")

            delta = newFit[counter] - bestFit

            # Update bestFit
            bestFit = newFit[counter]

            # Store l1norm
            l1Fit[counter] = MixingMatrix.l1norm(U @ A)

            # Store the unitary matrix
            # Us.append(U)

            # Check convergence
            if delta > np.sqrt(np.finfo(float).eps):
                if verbose:
                    print("Variable exchange optimization did not improve.")
            elif delta > -1e-5:
                if verbose:
                    print(
                        f"Variable exchange optimization converged within {counter} iterations."
                    )
                break

            # Update phase matrix
            Phases = np.exp(1j * np.angle(U @ A))

            if counter == maxIter - 1:
                print(
                    f"WARNING: Variable exchange optimization did not converge within "
                    f"{maxIter} iterations."
                )

        return U

    @staticmethod
    def procrustes(A, B):
        """
        Compute nearest orthogonal matrix U (in the Frobenius norm) via SVD such
        that it minimizes U*A - B.

        Parameters:
            A (numpy.array): First matrix
            B (numpy.array): Second matrix

        Returns:
            U (numpy.array): Unitary matrix that minimizes (U*A - B) in the Frobenius norm
        """
        # Compute SVD of B @ A.T
        W, _, V = np.linalg.svd(B @ A.conj().T)

        # Compute orthogonal polar factor of (B @ A.T)
        U = W @ V

        return U

    @staticmethod
    def smooth(C, C_plus):
        """
        Compute nearest UNITARY matrix U (in the Frobenius norm) via Procrustes
        solution such that (U*C_plus - C) is minimized. The output is the updated
        matrix C_n = U*C_plus.

        Parameters:
            A (numpy.array): Mixing matrix [Channels x Channels]
            B (numpy.array): Another mixing matrix [Channels x Channels]

        Returns:
            C_smooth (numpy.array): Smooth (w.r.t. neighbor C) mixing matrix [Channels x Channels]
        """
        U = MixingMatrix.procrustes(C_plus, C)
        C_smooth = U @ C_plus

        return C_smooth

    @staticmethod
    def balance_perserving_smoothness(C):
        """
        Compute balance Cbs = U * C with high l1(U * C) for U UNITARY while
        preserving smoothness (=single iteration of balance.m with closest
        phases of C).

        Parameters:
            C (numpy.array): Mixing matrix [Channels x Channels]

        Returns:
            Cbs (numpy.array): Smoothness-preserving balance mixing matrix [Channels x Channels]
        """
        # Closest phases (phases of C)
        B = np.exp(1j * np.angle(C, deg=False))

        # Procrustes solution
        U = MixingMatrix.procrustes(C, B)

        # Smoothing-preserving balanced matrix
        Cbs = U @ C

        return Cbs

    @staticmethod
    def l1norm(X):
        """
        Normalized l1 norm (element-wise)
        For ||X||_F = 1, the norm is in [0,1].
        """
        N = X.shape[0]

        return np.sum(np.abs(X)) / (N * np.sqrt(N))



import numpy as np
from scipy.special import jv


class Parameters:
    def __init__(self, mic_positions, sc_type, sample_frequency, nfft=1024, c=340, speed=20, direction=60):
        self.mic_positions = mic_positions  # Microphone positions [M x 3]
        # Spatial coherence model: 'corcos', 'spherical', or 'cylindrical'
        self.sc_type = sc_type.lower()
        self.sample_frequency = sample_frequency  # Sample frequency (Hz)
        self.nfft = nfft  # FFT length
        self.c = c  # Sound velocity (m/s)
        self.speed = speed  # Wind speed (m/s)
        self.direction = direction  # Wind direction (degrees)

    def copy(self):
        return Parameters(
            mic_positions=self.mic_positions.copy(),
            sc_type=self.sc_type,
            sample_frequency=self.sample_frequency,
            nfft=self.nfft,
            c=self.c,
            speed=self.speed,
            direction=self.direction
        )


class CoherenceMatrix:
    def __init__(self, params: Parameters):
        self.params = params
        self.matrix = self.generate_target_coherence()

    def generate_target_coherence(self):
        """
        Generate the analytical target coherence based on the model
        (e.g., diffuse spherical/cylindrical, Corcos), the sensor positions, 
        and the FFT length. Valid for an arbitrary 3D-array geometry.

        Returns:
            DC (np.array): Target coherence matrix [Channels x Channels x nfft/2+1]
        """

        # Angular frequency vector
        ww = 2 * np.pi * self.params.sample_frequency * \
            np.arange(self.params.nfft // 2 + 1) / self.params.nfft

        # Matrix of position vectors
        rr = self.params.mic_positions[:, None, :] - \
            self.params.mic_positions[None, :, :]

        # Matrix of inter-sensor distances
        d = np.linalg.norm(rr, axis=2)

        if self.params.sc_type == 'corcos':
            # Desired wind speed and direction
            Ud = self.params.speed / 3.6  # Wind speed [m/s]
            theta = np.deg2rad(self.params.direction)  # Angle in radians

            # Rotation matrix for wind direction
            R = np.array([[np.cos(theta), -np.sin(theta), 0],
                          [np.sin(theta), np.cos(theta), 0],
                          [0, 0, 1]])
            yy = np.array([0, 1, 0])  # y axis = true North
            u = np.dot(yy, R)  # Wind direction unit vector
            # Wind direction unit perpendicular vector
            u_p = np.array([u[1], -u[0], 0])

            # Coherence parameters
            alpha1, alpha2 = -0.125, -0.7
            alpha__ = alpha1 * np.abs(np.sum(u[None, None, :] * rr, axis=2)) \
                + alpha2 * np.abs(np.sum(u_p[None, None, :] * rr, axis=2))
            im__ = np.sum(u[None, None, :] * rr, axis=2)
            U = 0.8 * Ud  # Convective turbulence speed
            AA = np.einsum('ij,k->ijk', alpha__ - 1j * im__, ww)
            DC = np.exp(AA / U)

        elif self.params.sc_type == 'spherical':
            DC = np.sinc(d[:, :, None] * ww[None, None, :] /
                         (self.params.c * np.pi))

        elif self.params.sc_type == 'cylindrical':
            DC = jv(0,d[:, :, None] * ww[None, None, :] /
                         (self.params.c))

        else:
            raise ValueError(
                'Unknown coherence model. Please select "corcos", "spherical", or "cylindrical".')

        return DC
